from multiprocessing import Queue, Event as event
from multiprocessing.synchronize import Event
from pathlib import Path
from typing import Optional, Tuple, List

from fabric import Connection, Config
from invoke.exceptions import UnexpectedExit

from boba_fetch.rpc import UnixSocketRPCClient


class RemotePoller:
    def __init__(self, remote: str, tracker: UnixSocketRPCClient) -> None:
        self.remote = remote
        self.tracker = tracker
        # TODO: Make SSH config file configurable
        self._ssh_config = Config(runtime_ssh_path=(Path.home() / '.ssh' / 'config'))
        self._connection: Optional[Connection] = None

    @staticmethod
    def is_float(candidate: str) -> bool:
        return candidate.replace('.', '', 1).isdigit()

    @staticmethod
    def _process(cand: List[str]) -> Optional[Tuple[float, Path]]:
        if RemotePoller.is_float(cand[0]):
            return float(cand[0]), Path(cand[1])
        else:
            return None

    @property
    def connection(self) -> Connection:
        if self._connection is None or not self._connection.is_connected:
            self._connection = Connection(self.remote, config=self._ssh_config)
        return self._connection

    def ls(self, directory: Path) -> str:
        return f"find '{str(directory)}' -mindepth 1 -maxdepth 1 -printf '%T@ %p\n'"

    def find_new_files(self, directory: Path) -> List[Tuple[float, Path]]:
        try:
            results = self.connection.run(self.ls(directory), hide=True).stdout.split('\n')
            files = [RemotePoller._process(line.split(' ', 1)) for line in results if line]
            return [f for f in files if f is not None]
        except OSError as exc:
            print(f"Unable to connect to {self.remote}: {exc}")
        except UnexpectedExit as exc:
            cmd = self.ls(directory).replace('\n', '\\n')
            print(f"Unable to execute `{cmd}` on {self.remote}: {exc}")
        return []

    def poll(self, delay: float, directory: Path, queue: Queue, halt_flag: Optional[Event] = None) -> None:
        halt_flag = halt_flag or event()
        while not halt_flag.is_set():
            for mtime, new_file in self.find_new_files(directory):
                if not self.tracker.newer_than(new_file.name, mtime) and not \
                  self.tracker.currently_working(new_file.name):
                    queue.put(new_file)
            halt_flag.wait(timeout=delay)



from typeguard import typeguard_ignore
from typeguard.importhook import install_import_hook

from boba_fetch.rpc import UnixSocketRPCServer

install_import_hook('boba')

typeguard_ignore(UnixSocketRPCServer.server)
typeguard_ignore(UnixSocketRPCServer.proxy)



from multiprocessing import Queue, Process, Event as event
from pathlib import Path
from time import sleep

from pytest import mark

from boba_fetch.poller import RemotePoller
from boba_fetch.tracker import Tracker

from tests.unit.utils import tempdir, generate_big_sparse_file


@mark.parametrize('candidate,expected', [
    ('1', True),
    ('1.0', True),
    ('r', False),
    ('1.0.', False)
])
def test_is_float(candidate, expected) -> None:
    assert RemotePoller.is_float(candidate) == expected


@mark.parametrize('candidate,expected', [
    (['a', 'b'], None),
    (['1.5', 'stuff'], (1.5, Path('stuff'))),
    (['a', 'b', 'c'], None),
    (['1.5', 'stuff', 'more?'], (1.5, Path('stuff'))),
])
def test_process(candidate, expected) -> None:
    assert RemotePoller._process(candidate) == expected


def test_basic_polling() -> None:
    # NOTE: This test requires localhost SSH access, there's no way around this
    with tempdir() as tmp_dir:
        remote_dir = tmp_dir / 'stuff'
        remote_dir.mkdir()
        generate_big_sparse_file(remote_dir / 'A.file', 1024 * 1024)
        generate_big_sparse_file(remote_dir / 'B.file', 1024 * 1024)
        generate_big_sparse_file(remote_dir / 'C.file', 1024 * 1024)
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            poller = RemotePoller('localhost', tracker)
            new_files = set(f for _, f in poller.find_new_files(remote_dir))
            assert remote_dir / 'A.file' in new_files
            assert remote_dir / 'B.file' in new_files
            assert remote_dir / 'C.file' in new_files


def test_polling_exec_error() -> None:
    # NOTE: This test requires localhost SSH access, there's no way around this
    with tempdir() as tmp_dir:
        remote_dir = tmp_dir / 'stuff'
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            poller = RemotePoller('localhost', tracker)
            assert not set(f for _, f in poller.find_new_files(remote_dir))


def test_polling_conn_error() -> None:
    # NOTE: This test requires localhost SSH access, there's no way around this
    with tempdir() as tmp_dir:
        remote_dir = tmp_dir / 'stuff'
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            poller = RemotePoller('localhosty', tracker)
            assert not set(f for _, f in poller.find_new_files(remote_dir))


def test_repeated_polling() -> None:
    with tempdir() as tmp_dir:
        remote_dir = tmp_dir / 'stuff'
        remote_dir.mkdir()
        generate_big_sparse_file(remote_dir / 'A.file', 1024)
        generate_big_sparse_file(remote_dir / 'B.file', 1024)
        generate_big_sparse_file(remote_dir / 'C.file', 1024)
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            queue: Queue = Queue()
            poller = RemotePoller('localhost', tracker)
            halt_flag = event()

            def queue_consume():
                for _ in range(3):
                    new_file = queue.get()
                    tracker.start(new_file.name)
                    assert new_file in {remote_dir / 'A.file', remote_dir / 'B.file', remote_dir / 'C.file'}
                    tracker.mark_complete(new_file.name)
                sleep(0.2)
                (remote_dir / 'C.file').touch()
                new_file = queue.get()
                assert new_file == remote_dir / 'C.file'
                tracker.mark_complete(new_file.name)
                sleep(0.2)
                halt_flag.set()
            proc = Process(target=queue_consume)
            proc.start()
            poller.poll(0.1, remote_dir, queue, halt_flag=halt_flag)
            proc.join(timeout=1)



from threading import Thread

from boba_fetch.process import SafeProcess

from tests.unit.utils import tempdir


REFUSES_TO_TERMINATE = '''
from signal import SIGTERM, signal

def ignore_sigterm(_, __):
    pass

signal(SIGTERM, ignore_sigterm)

while 42:
    print('test')
'''


def test_uncooperative_process_termination() -> None:
    with tempdir() as tmp_dir:
        script = tmp_dir / 'uncooperative.py'
        script.open('w+').write(REFUSES_TO_TERMINATE)
        proc = SafeProcess(['python', str(script)])
        assert proc.terminate() is None
        for i, line in enumerate(proc.run()):
            assert line == 'test'
            if i == 5:
                proc.terminate(1)
        assert proc.returncode == -9


def test_process_termination() -> None:
    proc = SafeProcess(['sleep', '10'])

    def run():
        for line in proc.run():
            assert not line
    thread = Thread(target=run)
    thread.start()
    proc.started.wait()
    proc.terminate(timeout=1)
    thread.join()
    assert proc.returncode == -15



from pathlib import Path
from time import time
from typing import Optional, Dict, List

from boba_fetch.rpc import UnixSocketRPCServer


class Tracker(UnixSocketRPCServer):
    def __init__(self, directory: Path, check_in_timeout: float = 5.0) -> None:
        self.working: Dict[str, float] = {}
        self.directory = directory
        self.timeout = check_in_timeout
        if not self.directory.is_dir():
            self.directory.mkdir(parents=True, exist_ok=True)
        UnixSocketRPCServer.__init__(self)

    @UnixSocketRPCServer.serve
    def completed_at(self, name: str) -> float:
        if (self.directory / name).is_file():
            contents = (self.directory / name).open().read().strip()
            return float(contents) if contents.replace('.', '', 1).isdigit() else 0.0
        else:
            return 0.0

    @UnixSocketRPCServer.serve
    def mark_complete(self, name: str, timestamp: Optional[float] = None) -> None:
        (self.directory / name).open('w+').write(str(timestamp or time()))
        self.not_working(name)

    @UnixSocketRPCServer.serve
    def newer_than(self, name: str, timestamp: float) -> bool:
        return self.completed_at(name) >= timestamp

    @UnixSocketRPCServer.serve
    def start(self, name: str) -> None:
        self.check_in(name)
        (self.directory / name).open('w+').write('')

    @UnixSocketRPCServer.serve
    def failed(self, name: str) -> None:
        (self.directory / name).unlink(missing_ok=True)
        self.not_working(name)

    def not_working(self, name: str) -> None:
        if name in self.working:
            del self.working[name]

    def clean(self) -> None:
        now = time()
        for name in [n for n, c in self.working.items() if now - c >= self.timeout]:
            self.failed(name)

    @UnixSocketRPCServer.serve
    def check_in(self, name: str) -> None:
        self.clean()
        self.working[name] = time()

    @UnixSocketRPCServer.serve
    def currently_working(self, name: str) -> bool:
        return name in self.working and time() - self.working[name] <= self.timeout

    @UnixSocketRPCServer.serve
    def unfinished(self) -> List[str]:
        self.clean()
        return [n for n in self.working]



from dataclasses import dataclass
from multiprocessing import Process, Event as event, Queue
from multiprocessing.synchronize import Event
from pathlib import Path
from signal import SIGTERM, SIGINT, signal
from tempfile import mkdtemp
from typing import Optional, List
from warnings import simplefilter

import click
from click_path import PathlibPath
import cryptography

from boba_fetch.poller import RemotePoller
from boba_fetch.rpc import UnixSocketRPCClient
from boba_fetch.syncer import Syncer, WatchParams
from boba_fetch.tracker import Tracker


simplefilter('ignore', cryptography.utils.CryptographyDeprecationWarning)


@dataclass
class BobaConfig:
    remote_host: str
    remote_dir: Path
    local_dir: Path
    staging_dir: Path
    tracking_dir: Path
    num_sync_workers: int
    delete_src: bool
    execute_on_complete: Optional[str]
    polling_delay: float

    def to_watch_params(self, queue: Queue) -> WatchParams:
        return WatchParams(self.remote_host, self.local_dir, queue, self.delete_src,
                           self.execute_on_complete)


def run_poller(config: BobaConfig,
               tracker: UnixSocketRPCClient,
               queue: Queue,
               halt_flag: Event):
    poller = RemotePoller(config.remote_host, tracker)
    poller.poll(config.polling_delay, config.remote_dir, queue, halt_flag)


def start_poller(config: BobaConfig,
                 tracker: UnixSocketRPCClient,
                 queue: Queue,
                 halt_flag: Event) -> Process:
    poller_proc = Process(target=run_poller, args=(config, tracker, queue, halt_flag))
    poller_proc.start()
    return poller_proc


def run_syncer(config: BobaConfig,
               tracker: UnixSocketRPCClient,
               queue: Queue,
               halt_flag: Event):
    syncer = Syncer(tracker, halt_flag, config.staging_dir)
    syncer.watch_queue(config.to_watch_params(queue))


def start_syncers(config: BobaConfig,
                  tracker: UnixSocketRPCClient,
                  queue: Queue,
                  halt_flag: Event) -> List[Process]:
    syncer_procs = [Process(target=run_syncer,
                            args=(config, tracker,
                                  queue, halt_flag)) for _ in range(config.num_sync_workers)]
    for proc in syncer_procs:
        proc.start()
    return syncer_procs


@click.command()
@click.option('-r', '--remote', required=True,
              help='The remote host to poll for file updates. Must be SSH accessible via public key '
                   'and configured in your ~/.ssh/config')
@click.option('-s', '--src-dir', required=True, type=PathlibPath(exists=False),
              help='The directory on the remote host to check for files to download')
@click.option('-d', '--dst-dir', required=True, type=PathlibPath(exists=True, file_okay=False,
                                                                 dir_okay=True, writable=True,
                                                                 resolve_path=True),
              help='Local directory to sync the files to from the remote host')
@click.option('-g', '--staging-dir', default=None, type=PathlibPath(exists=True, file_okay=False,
                                                                    dir_okay=True, writable=True,
                                                                    resolve_path=True),
              help='Staging directory where files are saved, defaults to randomized temporary directory')
@click.option('-t', '--tracking-dir', required=True, type=PathlibPath(exists=True, file_okay=False,
                                                                      dir_okay=True, writable=True,
                                                                      resolve_path=True),
              help='Tracking directory where we keep track of files that have been synced')
@click.option('-n', '--num-sync-workers', default=3, type=int,
              help='Number of sync workers to use for downloads, defaults to 3')
@click.option('--delete-src', is_flag=True, default=False, help='Indicates to whether to delete the source files on '
                                                                'the remote system')
@click.option('-e', '--execute-on-complete', default=None, help='A command to execute every time a sync is complete')
@click.option('-p', '--polling-delay', default=30.0, type=float,
              help='Number of seconds to wait between polls of the remote directory, defaults to 30')
def main(remote: str,
         src_dir: Path,
         dst_dir: Path,
         staging_dir: Optional[Path],
         tracking_dir: Path,
         num_sync_workers: int,
         delete_src: bool,
         execute_on_complete: Optional[str],
         polling_delay: float):
    queue: Queue = Queue()
    halt_flag = event()
    config = BobaConfig(remote, src_dir, dst_dir, staging_dir or Path(mkdtemp()),
                        tracking_dir, num_sync_workers, delete_src, execute_on_complete,
                        polling_delay)
    with Tracker(config.tracking_dir).proxy() as tracker:
        syncer_procs = start_syncers(config, tracker, queue, halt_flag)
        poller_proc = start_poller(config, tracker, queue, halt_flag)

        def handle_signal(_, __):
            halt_flag.set()
        signal(SIGTERM, handle_signal)
        signal(SIGINT, handle_signal)
        poller_proc.join()
        print("Poller shutdown complete.")
        for sync_proc in syncer_procs:
            sync_proc.join()
        print("Syncer shutdown complete.")
        print("Done.")



from contextlib import contextmanager
from http.client import HTTPConnection
from multiprocessing import Process, Event as event
from multiprocessing.synchronize import Event
from pathlib import Path
from random import choice
from socket import socket, AF_UNIX, SOCK_STREAM
from socketserver import UnixStreamServer
from string import ascii_letters, digits
from tempfile import mkdtemp
from typing import Optional, Any, Union, Tuple, Dict, Set, Iterator
from xmlrpc.client import Transport, ServerProxy
from xmlrpc.server import SimpleXMLRPCDispatcher, SimpleXMLRPCRequestHandler


class _UnixStreamXMLRPCRequestHandler(SimpleXMLRPCRequestHandler):
    disable_nagle_algorithm = False

    def address_string(self) -> Any:
        return self.client_address  # pragma: no cover


class UnixStreamXMLRPCServer(UnixStreamServer, SimpleXMLRPCDispatcher):
    def __init__(self, addr: Path,
                 log_requests: bool = False,
                 allow_none: bool = True,
                 encoding: Optional[str] = None,
                 bind_and_activate: bool = True,
                 use_builtin_types: bool = True):
        self.logRequests = log_requests  # pylint: disable=C0103
        SimpleXMLRPCDispatcher.__init__(self, allow_none, encoding, use_builtin_types)
        UnixStreamServer.__init__(self, str(addr), _UnixStreamXMLRPCRequestHandler,
                                  bind_and_activate)


class _UnixStreamHTTPConnection(HTTPConnection):
    def connect(self) -> None:
        self.sock = socket(AF_UNIX, SOCK_STREAM)
        self.sock.connect(self.host)


class _UnixStreamTransport(Transport):
    def __init__(self, socket_path: Path):
        self.socket_path = socket_path
        super().__init__()

    def make_connection(self, host: Union[Tuple[str, Dict[str, str]], str]):
        return _UnixStreamHTTPConnection(str(self.socket_path))


class UnixSocketRPCClient(ServerProxy):
    def __init__(self, addr: Path, **kwargs: Any):
        transport = _UnixStreamTransport(addr)
        super().__init__("http://", transport=transport, **kwargs)


def _random_string(length) -> str:
    return ''.join([choice(ascii_letters + digits) for _ in range(length)])


class ServedFunctionRegistrar:
    def __init__(self) -> None:
        self.served_functions: Dict[str, Set[str]] = {}
        # self.served_functions: Set[str] = set()

    def __call__(self, method):
        class_name = method.__qualname__.split(".")[-2]
        if class_name not in self.served_functions:
            self.served_functions[class_name] = set()
        self.served_functions[class_name].add(method.__name__)

        def decorate(*args, **kwargs):
            return method(*args, **kwargs)  # pragma: no cover
        return decorate


class UnixSocketRPCServer():
    serve = ServedFunctionRegistrar()

    def __init__(self, socket_path: Optional[Path] = None):
        self.socket_path = socket_path or (Path(mkdtemp()) / (_random_string(16) + '.sock'))

    @property
    def served(self) -> Set[str]:
        return UnixSocketRPCServer.serve.served_functions[self.__class__.__name__]

    def serve_forever(self, started_flag: Optional[Event] = None):
        started_flag = started_flag or event()
        with UnixStreamXMLRPCServer(self.socket_path) as server:
            server.register_introspection_functions()
            for method_name in self.served:
                server.register_function(getattr(self, method_name), method_name)
            started_flag.set()
            server.serve_forever()

    def server_process(self, daemon: bool = True) -> Process:
        started = event()
        proc = Process(target=self.serve_forever, args=(started,), daemon=daemon)
        proc.start()
        started.wait()
        return proc

    @contextmanager
    def server(self, terminate_timeout: int = 1) -> Iterator[Path]:
        proc = self.server_process()
        try:
            yield self.socket_path
        finally:
            proc.terminate()
            proc.join(timeout=terminate_timeout)
            if proc.exitcode is None:
                proc.kill()
                proc.join()

    @contextmanager
    def proxy(self, terminate_timeout: int = 1) -> Iterator[UnixSocketRPCClient]:
        with self.server(terminate_timeout) as socket_path:
            yield UnixSocketRPCClient(socket_path)
